#!/usr/bin/env python3

import os
import sys
import fire
import getpass
import configparser
from os.path import expanduser
from .adfs_auth import authenticate
from .assume_role import ask_which_role_to_assume, assume_role
from .roles import action_url_on_validation_success, retrieve_roles_page


def cli(account=None, role=None, profile='default'):
    # Get the federated credentials from the user
    cached_netid = load_last_netid(aws_file('config'), profile)
    if cached_netid:
        net_id_prompt = 'BYU Net ID [{}]: '.format(cached_netid)
    else:
        net_id_prompt = 'BYU Net ID: '
    net_id = input(net_id_prompt) or cached_netid
    if "@byu.local" in net_id:
        print('@byu.local is not required')
        username = net_id
    else:
        username = '{}@byu.local'.format(net_id)
    password = getpass.getpass()
    print('')

    ####
    # Authenticate against ADFS with DUO MFA
    ####
    html_response, session, auth_signature, duo_request_signature = authenticate(username, password)

    # Overwrite and delete the credential variables, just for safety
    username = '##############################################'
    password = '##############################################'
    del username
    del password

    ####
    # Obtain the roles available to assume
    ####
    roles_page_url = action_url_on_validation_success(html_response)
    account_names, principal_roles, assertion, aws_session_duration = retrieve_roles_page(
        roles_page_url,
        html_response,
        session,
        auth_signature,
        duo_request_signature,
    )

    ####
    # Ask user which role to assume
    ####
    account_roles = ask_which_role_to_assume(account_names, principal_roles, account, role)

    ####
    # Assume roles and set in the environment
    ####
    for account_role in account_roles:
        aws_session_token = assume_role(account_role, assertion)

        # If assuming roles across all accounts, then use the account name as the profile name
        if account == 'all':
            profile = account_role.account_name

        check_for_aws_dir()
        write_to_cred_file(aws_file('creds'), aws_session_token, profile)
        write_to_config_file(aws_file('config'), net_id, 'us-west-2', profile)

        print("Now logged into {}@{}".format(account_role.role_name, account_role.account_name))


def main():
    if not sys.version.startswith('3.6'):
        sys.stderr.write("byu_awslogin requires python 3.6\n")
        sys.exit(-1)
    fire.Fire(cli)


def aws_file(file_type):
    if file_type == 'creds':
        return "{}/.aws/credentials".format(expanduser('~'))
    else:
        return "{}/.aws/config".format(expanduser('~'))


def open_config_file(file):
    config = configparser.ConfigParser()
    config.read(file)
    return config


def check_for_aws_dir(directory="{}/.aws".format(expanduser('~'))):
    if not os.path.exists(directory):
        os.makedirs(directory)


def write_to_cred_file(file, aws_session_token, profile):
    config = open_config_file(file)
    config[profile] = {
        'aws_access_key_id': aws_session_token['Credentials']['AccessKeyId'],
        'aws_secret_access_key': aws_session_token['Credentials']['SecretAccessKey'],
        'aws_session_token': aws_session_token['Credentials']['SessionToken']
    }
    with open(file, 'w') as configfile:
        config.write(configfile)


def write_to_config_file(file, net_id, region, profile):
    config = open_config_file(file)
    config[profile] = {
        'region': region, 'adfs_netid': net_id
    }
    with open(file, 'w') as configfile:
        config.write(configfile)


def load_last_netid(file, profile):
    config = open_config_file(file)
    if config.has_section(profile) and config.has_option(profile, 'adfs_netid'):
        return config[profile]['adfs_netid']
    else:
        return ''


if __name__ == '__main__':
    main()



import base64
from bs4 import BeautifulSoup
import lxml.etree as ET
import re
import requests

default_session_duration = 3600


_headers = {
    'Accept-Language': 'en',
    'User-Agent': 'Mozilla/5.0 (Windows NT 6.1; WOW64; Trident/7.0; rv:11.0) like Gecko',
    'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8',
    'Accept': 'text/plain, */*; q=0.01',
}

def extract(saml_page_html_soup):
    # Retrieve Base64-encoded SAML assertion from form SAMLResponse input field
    saml_assertion = saml_page_html_soup.find('form', {"name":"hiddenform"}).find('input', {"name":"SAMLResponse"})['value']

    # If we did not get an error, but also do not have an assertion,
    # then the user needs to authenticate
    if not saml_assertion:
        print("Unknown error: No assertion returned")
        exit(1)

    # Parse the returned assertion and extract the authorized roles
    saml = ET.fromstring(base64.b64decode(saml_assertion))

    # Find all roles offered by the assertion
    raw_roles = saml.findall(
        './/{*}Attribute[@Name="https://aws.amazon.com/SAML/Attributes/Role"]/{*}AttributeValue'
    )
    aws_roles = [element.text.split(',') for element in raw_roles]

    # Note the format of the attribute value is provider_arn, role_arn
    principal_roles = [role for role in aws_roles if ':saml-provider/' in role[0]]

    aws_session_duration = default_session_duration
    # Retrieve session duration
    for element in saml.findall('.//{*}Attribute[@Name="https://aws.amazon.com/SAML/Attributes/SessionDuration"]/{*}AttributeValue'):
        aws_session_duration = int(element.text)
        
    account_names = _get_account_names(saml_assertion)

    return account_names, principal_roles, saml_assertion, aws_session_duration


def action_url_on_validation_success(login_response_html_soup):
    options_form = login_response_html_soup.find('form', id='options')
    return options_form['action']


def retrieve_roles_page(roles_page_url, login_response_html_soup, session, auth_signature, duo_request_signature):
    context = _context(login_response_html_soup)
    signed_response = '{}:{}'.format(auth_signature, _app(duo_request_signature))
    roles_page_response = session.post(
        roles_page_url,
        verify=True,
        headers=_headers,
        allow_redirects=True,
        data={
            'AuthMethod': 'DuoAdfsAdapter',
            'Context': context,
            'sig_response': signed_response,
        }
    )

    if roles_page_response.status_code != 200:
        raise RuntimeError(
            u'Issues during redirection to aws roles page. The error response {}'.format(
                roles_page_response
            )
        )

    roles_page_html_soup = BeautifulSoup(roles_page_response.text, 'lxml')
    return extract(roles_page_html_soup)


def _get_account_names(saml_assertion):
    saml_url = "https://signin.aws.amazon.com:443/saml"
    headers = {
        'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8'
    }
    response = requests.post(saml_url, headers=headers, data={
        'SAMLResponse': saml_assertion
    })
    response.raise_for_status()
    html_response = ET.fromstring(response.text, ET.HTMLParser())
    account_names = {}
    for element in html_response.findall('.//div[@class="saml-account-name"]'):
        account_id = element.text.split(' ')[2].replace('(', '').replace(')', '')
        account_name = element.text.split(' ')[1]
        account_names[account_id] = account_name
    
    return account_names
    

def _app(request_signature):
    app_pattern = re.compile(".*(APP\|[^:]+)")
    m = app_pattern.search(request_signature)
    return m.group(1)


def _context(login_response_html_soup):
    context_input = login_response_html_soup.find('input', id='context')
    return context_input['value']



import pytest
import os.path
import configparser
from byu_awslogin import index


@pytest.fixture
def aws_config_file(tmpdir):
    return str(tmpdir.dirpath(".aws/config"))


@pytest.fixture
def aws_cred_file(tmpdir):
    return str(tmpdir.dirpath(".aws/credentials"))


@pytest.fixture
def read_config_file(file):
    config = configparser.ConfigParser()
    config.read(file)
    return config


@pytest.fixture
def profile():
    return 'default'

@pytest.fixture
def fake_config_file(tmpdir):
    dir = tmpdir.dirpath(".aws/")
    file = dir + 'config'
    if not os.path.exists(dir):
        os.makedirs(dir)
    config = configparser.ConfigParser()
    config['default'] = {'region': 'fake-region-1', 'adfs_netid': 'fake_netid'}
    with open(file, 'w') as write_file:
        config.write(write_file)
    return file


def test_open_config_file(aws_config_file):
    assert index.open_config_file(aws_config_file)


def test_aws_file():
    assert index.aws_file('config')


def test_check_for_aws_dir(tmpdir):
    dir = tmpdir.dirpath(".aws")
    index.check_for_aws_dir(dir)
    assert os.path.exists(dir)


def test_write_to_cred_file(aws_cred_file, profile):
    aws_session_token = {'Credentials': {'AccessKeyId': 'keyid', 'SecretAccessKey': 'secretkey', 'SessionToken': 'sessiontoken'}}
    index.write_to_cred_file(aws_cred_file, aws_session_token, profile)
    config = read_config_file(aws_cred_file)
    assert config['default'] == {'aws_access_key_id': 'keyid',
                                 'aws_secret_access_key': 'secretkey',
                                 'aws_session_token': 'sessiontoken'
                                 }


def test_write_to_config_file(aws_config_file, profile):
    net_id = 'fake_netid'
    region = 'fakeRegion'
    index.write_to_config_file(aws_config_file, net_id, region, profile)
    config = read_config_file(aws_config_file)
    assert config['default'] == {'region': 'fakeRegion', 'adfs_netid': 'fake_netid'}


def test_load_last_netid(fake_config_file, profile):
    assert index.load_last_netid(str(fake_config_file), profile) == 'fake_netid'


@pytest.mark.skip(reason="not sure how to test this yet")
def test_cli():
    pass



import boto3
import botocore
from botocore import client


class AccountRole:

    def __init__(self, account_name, role_name, role_arn, principal_arn):
        self.account_name = account_name
        self.role_name = role_name
        self.role_arn = role_arn
        self.principal_arn = principal_arn


def ask_which_role_to_assume(account_names, principal_roles, account_name=None, role_name=None):
    roles_by_account = __get_roles_by_account(account_names, principal_roles)

    # User wants to assume a role on all accounts
    if account_name == 'all':
        return __get_all_account_roles(roles_by_account, role_name)
    # User wants to assume a role in a single account
    else:
        return __get_single_account_role(roles_by_account, account_name, role_name)


def assume_role(account_role, samlAssertion):
    conn = boto3.client('sts', config=client.Config(signature_version=botocore.UNSIGNED))
    aws_session_token = conn.assume_role_with_saml(
        RoleArn=account_role.role_arn,
        PrincipalArn=account_role.principal_arn,
        SAMLAssertion=samlAssertion,
        DurationSeconds=3600,
    )

    return aws_session_token


def __get_all_account_roles(roles_by_account, role_name):
    if not role_name:
        print("You must speicify a role using the --role option when assuming a role accross all accounts")
        exit(1)
    account_roles = []
    for account_name in roles_by_account:
        if role_name in roles_by_account[account_name]:
            account_roles.append(roles_by_account[account_name][role_name])
        else:
            print("Will not try to assume role '{}' in account '{}' because you don't have access to that role".format(
                role_name, account_name))
    return account_roles


def __get_single_account_role(roles_by_account, account_name, role_name):
    # Prompt the user for account (if not already provided)
    if not account_name or not roles_by_account[account_name]:
        account_names = list(roles_by_account.keys())
        account_names.sort()
        account_name = __prompt_for_account(account_names)
    # Prompt the user for role (if not already provided)
    if not role_name or not roles_by_account[account_name][role_name]:
        role_names = __get_role_names_for_account(roles_by_account, account_name)
        role_names.sort()
        if len(role_names) == 1:
            role_name = role_names[0]
        else:
            role_name = __prompt_for_role(account_name, role_names)

    return [roles_by_account[account_name][role_name]]  # TODO - Return real list at some point


def __get_role_names_for_account(roles_by_account, account_name):
    role_names = []
    for role_name in roles_by_account[account_name]:
        role_names.append(role_name)
    return role_names

    
def __get_roles_by_account(account_names, principal_roles):
    roles_by_account = {}
    for (principal_arn, role_arn) in principal_roles:
        account_id = role_arn.split(':')[4]
        account_name = account_names[account_id]
        role_name = role_arn.split(':')[5].split('/')[1]

        roles_by_account.setdefault(account_name, {})[role_name] = AccountRole(account_name, role_name, role_arn, principal_arn)
    return roles_by_account


def __prompt_for_account(account_names):
    print('#--------------------------------------------------')
    print('# To which account would you like to login?')
    print('#--------------------------------------------------')
    
    for index, account_name in enumerate(account_names):
        print('{} {}'.format(str(index).rjust(2), account_name))
        
    while True:
        choice = input('Select account: ')
        try:
            return account_names[int(choice)]
        except:
            maximum = len(account_names) - 1
            print('Please enter an integer between 0 and {}'.format(maximum))


def __prompt_for_role(account_name, role_names):
    print('#--------------------------------------------------')
    print('# You have access to the following roles in {}'.format(account_name))
    print('# Which role would you like to assume?')
    print('#--------------------------------------------------')
    
    for index, role_name in enumerate(role_names):
        print('{} {}'.format(str(index).rjust(2), role_name))
    
    while True:
        choice = input('Select role: ')
        try:
            return role_names[int(choice)]
        except:
            maximum = len(role_names) - 1
            print('Please enter an integer between 0 and {}'.format(maximum))
    


#!/usr/bin/env python3

import os
import sys
import subprocess
from setuptools import setup, find_packages

if not sys.version.startswith('3.6'):
    sys.stderr.write("byu_awslogin requires python 3.6\n")
    sys.exit(-1)

subprocess.check_call('pandoc --from=markdown_github --to=rst --output=README.rst README.md', shell=True)
with open("README.rst") as rm_file:
    long_description = rm_file.read()
os.remove('README.rst')

def get_requirements():
    with open('requirements.txt') as obj:
        lines = [dep for dep in obj.read().split('\n') if dep]
        return lines

VERSION = ''
with open('VERSION') as verobj:
    VERSION = verobj.read()

setup(name='byu_awslogin',
      version=VERSION,
      description="An aws-adfs spinoff that fits BYU's needs",
      long_description=long_description,
      author='BYU OIT Application Development',
      author_email='it@byu.edu',
      url='https://github.com/byu-oit/awslogin',
      packages=find_packages(),
      license="Apache 2",
      install_requires=get_requirements(),
      zip_safe=True,
      entry_points={
          'console_scripts': ['awslogin=byu_awslogin.index:main']
      },
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: System Administrators',
          'Intended Audience :: Education',
          'License :: OSI Approved :: Apache Software License',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: Microsoft :: Windows',
          'Operating System :: POSIX',
          'Programming Language :: Python :: 3 :: Only',
          'Natural Language :: English',
          'Topic :: Utilities'
      ]
)



#
# This file was adapted from https://github.com/venth/aws-adfs. Thanks to https://github.com/venth for his work on
# figuring this out
#

import requests
from bs4 import BeautifulSoup
import re
import lxml.etree as ET
from urllib.parse import urlparse, parse_qs

# idpentryurl: The initial url that starts the authentication process.
adfs_entry_url = 'https://awslogin.byu.edu:443/adfs/ls/IdpInitiatedSignOn.aspx?loginToRp=urn:amazon:webservices'

def authenticate(username, password):
    # Initiate session handler
    session = requests.Session()

    # Get the ADFS sign-in page HTML
    login_page_response = session.get(adfs_entry_url, verify=True)

    # Parse the response and extract all the necessary values
    # in order to build a dictionary of all of the form values the IdP expects
    login_html_soup = BeautifulSoup(login_page_response.text, "lxml")
    auth_payload = _get_auth_payload(login_html_soup, username, password)

    # From the form action for the login form, build the URL used to submit the login request
    adfs_form_submit_url = _get_login_submit_url(login_html_soup)

    # Login with the ADFS credentials
    login_response = session.post(
        adfs_form_submit_url, data=auth_payload, verify=True)

    login_response_html_soup = BeautifulSoup(login_response.text, 'lxml')

    # Check that authentication succeeded. Exit with error if it didn't
    _check_adfs_authentication_success(login_response_html_soup)

    # Perform DUO MFA
    auth_signature, duo_request_signature = _authenticate_duo(login_response_html_soup, True, session)

    return login_response_html_soup, session, auth_signature, duo_request_signature

_headers = {
    'Accept-Language': 'en',
    'User-Agent': 'Mozilla/5.0 (Windows NT 6.1; WOW64; Trident/7.0; rv:11.0) like Gecko',
    'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8',
    'Accept': 'text/plain, */*; q=0.01',
}

def _get_auth_payload(login_html_soup, username, password):
    auth_payload = {}
    for inputtag in login_html_soup.find_all(re.compile('(INPUT|input)')):
        name = inputtag.get('name', '')
        value = inputtag.get('value', '')
        if "user" in name.lower():
            # Make an educated guess that this is the right field for the username
            auth_payload[name] = username
        elif "pass" in name.lower():
            # Make an educated guess that this is the right field for the password
            auth_payload[name] = password
        else:
            # Simply populate the parameter with the existing value (picks up hidden fields in the login form)
            auth_payload[name] = value
    return auth_payload


def _get_login_submit_url(login_html_soup):
    parsed_adfs_login_url = urlparse(adfs_entry_url)
    adfs_form_submit_url = parsed_adfs_login_url.scheme + "://" + parsed_adfs_login_url.netloc
    for inputtag in login_html_soup.find_all(re.compile('(FORM|form)')):
        action = inputtag.get('action')
        loginid = inputtag.get('id')
        if (action and loginid == "loginForm"):
            adfs_form_submit_url += action
    return adfs_form_submit_url


def _check_adfs_authentication_success(login_response_html_soup):
    login_form_tag = login_response_html_soup.find('form', id='loginForm')
    if login_form_tag: # Login form present means the authentication failed
        auth_error = login_form_tag.find('span', id='errorText')
        print(auth_error.string)
        exit(1)

def _authenticate_duo(duo_page_html_soup, roles_page_url, session):
    duo_host = _duo_host(duo_page_html_soup)
    duo_request_signature = _duo_request_signature(duo_page_html_soup)

    print("Sending request for authentication")
    (sid, preferred_factor, preferred_device), initiated = _initiate_authentication(
        duo_host,
        duo_request_signature,
        roles_page_url,
        session
    )
    if initiated:
        transaction_id = _begin_authentication_transaction(
            duo_host,
            sid,
            preferred_factor,
            preferred_device,
            session
        )

        print("Waiting for additional authentication")
        _verify_that_code_was_sent(
            duo_host,
            sid,
            transaction_id,
            session
        )
        auth_signature = _authentication_result(
            duo_host,
            sid,
            transaction_id,
            session,
        )

        return auth_signature, duo_request_signature
    else:
        raise RuntimeError("DUO Transaction Not Initiated")


def _authentication_result(
        duo_host,
        sid,
        duo_transaction_id,
        session
):
    status_for_url = "https://{}/frame/status".format(duo_host)
    response = session.post(
        status_for_url,
        verify=True,
        headers=_headers,
        data={
            'sid': sid,
            'txid': duo_transaction_id
        }
    )

    if response.status_code != 200:
        raise RuntimeError(
            u'Issues during retrieval of a code entered into '
            u'the device. The error response {}'.format(
                response
            )
        )

    json_response = response.json()
    if json_response['stat'] != 'OK':
        raise RuntimeError(
            u'There was an issue during retrieval of a code entered into the device.'
            u' The error response: {}'.format(
                response.text
            )
        )

    if json_response['response']['status_code'] != 'allow':
        raise RuntimeError(
            u'There was an issue during retrieval of a code entered into the device.'
            u' The error response: {}'.format(
                response.text
            )
        )

    auth_signature = response.json()['response']['cookie']
    return auth_signature


def _verify_that_code_was_sent(duo_host, sid, duo_transaction_id, session):
    status_for_url = "https://{}/frame/status".format(duo_host)
    response = session.post(
        status_for_url,
        verify=True,
        headers=_headers,
        data={
            'sid': sid,
            'txid': duo_transaction_id
        }
    )

    if response.status_code != 200:
        raise RuntimeError(
            u'Issues during sending code to the devide. The error response {}'.format(
                response
            )
        )

    json_response = response.json()
    if json_response['stat'] != 'OK':
        raise RuntimeError(
            u'There was an issue during sending code to the device. The error response: {}'.format(
                response.text
            )
        )

    if json_response['response']['status_code'] != 'pushed':
        raise RuntimeError(
            u'There was an issue during sending code to the device. The error response: {}'.format(
                response.text
            )
        )


def _tx(request_signature):
    tx_pattern = re.compile("(TX\|[^:]+):APP.+")
    m = tx_pattern.search(request_signature)
    return m.group(1)


def _initiate_authentication(duo_host, duo_request_signature, roles_page_url, session):
    prompt_for_url = 'https://{}/frame/web/v1/auth'.format(duo_host)
    parent = "{}{}".format(
        roles_page_url,
        "&java_version="
        "&flash_version="
        "&screen_resolution_width=1280"
        "&screen_resolution_height=800"
        "&color_depth=24"
    )
    response = session.post(
        prompt_for_url,
        verify=True,
        headers={
            'Host': duo_host,
            'User-Agent': "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.12; rv:52.0) Gecko/20100101 Firefox/52.0",
            'Accept': "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
            'Accept-Language': "en-US,en;q=0.5",
            'Accept-Encoding': "gzip, deflate, br",
            'DNT': "1",
            'Upgrade-Insecure-Requests': "1",
            'Content-Type': "application/x-www-form-urlencoded",
        },
        allow_redirects=True,
        params={
            'tx': _tx(duo_request_signature),
            'parent': parent,
            'v': '2.3',
        },
        data={
            'parent': parent,
            'java_version': '',
            'flash_version': '22.0.0.209',
            'screen_resolution_width': '1280',
            'screen_resolution_height': '800',
            'color_depth': '24',
        }
    )

    if response.status_code != 200 or response.url is None:
        return None, False

    o = urlparse(response.url)
    query = parse_qs(o.query)

    if 'sid' not in query:
        return None, False

    sid = query['sid']
    html_response = ET.fromstring(response.text, ET.HTMLParser())
    preferred_factor = _preferred_factor(html_response)
    preferred_device = _preferred_device(html_response)
    return (sid, preferred_factor, preferred_device), True


def _preferred_factor(html_response):
    preferred_factor_query = './/input[@name="preferred_factor"]'
    element = html_response.find(preferred_factor_query)
    return element.get('value')


def _preferred_device(html_response):
    preferred_device_query = './/input[@name="preferred_device"]'
    element = html_response.find(preferred_device_query)
    return element.get('value')


def _begin_authentication_transaction(duo_host, sid, preferred_factor, preferred_device, session):
    prompt_for_url = "https://{}/frame/prompt".format(duo_host)
    response = session.post(
        prompt_for_url,
        verify=True,
        headers=_headers,
        data={
            'sid': sid,
            'factor': preferred_factor,
            'device': preferred_device,
            'out_of_date': ''
        }
    )

    if response.status_code != 200:
        raise RuntimeError(
            u'Issues during beginning of the authentication process. The error response {}'.format(
                response
            )
        )

    json_response = response.json()
    if json_response['stat'] != 'OK':
        raise RuntimeError(
            u'Cannot begin authentication process. The error response: {}'.format(response.text)
        )

    return json_response['response']['txid']


def _duo_host(duo_page_html_soup):
    duo_script = duo_page_html_soup.find('form', id='duo_form').find_next_sibling('script').string
    duo_host_pattern = re.compile("'host': '([^']+)'")
    m = duo_host_pattern.search(duo_script)
    return m.group(1)

def _duo_request_signature(duo_page_html_soup):
    duo_script = duo_page_html_soup.find('form', id='duo_form').find_next_sibling('script').string
    duo_signature_pattern = re.compile("'sig_request': '([^']+)'")
    m = duo_signature_pattern.search(duo_script)
    return m.group(1)
